/**
 * Copyright (c) 2023 Raspberry Pi (Trading) Ltd.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */
#define BTSTACK_FILE__ "server.c"
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include "btstack.h"
#include "pico/cyw43_arch.h"
#include "pico/btstack_cyw43.h"
#include "pico/stdlib.h"
#include "ble/gatt-service/nordic_spp_service_server.h"
#include "picow_nus.h"
#include "server_common.h"

int le_notification_enabled;
static uint32_t counter = 0;
static char counter_string[30];
static uint8_t  counter_string_len;
static uint8_t  connect_flag;
static int led_on = true;
 
#define APP_AD_FLAGS 0x06
static uint8_t adv_data[] = {
    // Flags general discoverable, BR/EDR not supported
    0x02, BLUETOOTH_DATA_TYPE_FLAGS, APP_AD_FLAGS, 
    // Name
    0x06, BLUETOOTH_DATA_TYPE_COMPLETE_LOCAL_NAME, 'P', 'i', 'c', 'o', 'W',
    0x11, BLUETOOTH_DATA_TYPE_COMPLETE_LIST_OF_128_BIT_SERVICE_CLASS_UUIDS, 0x9e, 0xca, 0xdc, 0x24, 0xe, 0xe5, 0xa9, 0xe0, 0x93, 0xf3, 0xa3, 0xb5, 0x1, 0x0, 0x40, 0x6e,
};
static const uint8_t adv_data_len = sizeof(adv_data);

#define HEARTBEAT_PERIOD_MS 1000

static btstack_timer_source_t heartbeat;
hci_con_handle_t con_handle;
btstack_context_callback_registration_t send_request;
static btstack_packet_callback_registration_t hci_event_callback_registration;

static void heartbeat_handler(struct btstack_timer_source *ts) {
    if(con_handle != HCI_CON_HANDLE_INVALID)
    {
        counter++;
        counter_string_len = snprintf(counter_string,sizeof(counter_string),"BTstack counter %03u",counter);
        send_request.callback = &nordic_can_send;
        nordic_spp_service_server_request_can_send_now(&send_request, con_handle);
    }

     switch(connect_flag)
    {
        case 0:
            // Invert the led
            led_on = !led_on;
            cyw43_arch_gpio_put(CYW43_WL_GPIO_LED_PIN, led_on);
        break;
        case 1:
            // DO NOT Invert the led
            led_on = true;
            cyw43_arch_gpio_put(CYW43_WL_GPIO_LED_PIN, led_on);
        break;
    }
    // Restart timer
    heartbeat.process = &heartbeat_handler;
    btstack_run_loop_set_timer(&heartbeat, HEARTBEAT_PERIOD_MS);
    btstack_run_loop_add_timer(&heartbeat);

}

static void nordic_can_send(void * context){
    UNUSED(context);
    printf("SEND: %s\n",counter_string);
    nordic_spp_service_server_send(con_handle, (uint8_t*) counter_string, counter_string_len);
}

void stack_event_packet_handler(uint8_t packet_type, uint16_t channel, uint8_t *packet, uint16_t size) {
    UNUSED(size);
    UNUSED(channel);
    bd_addr_t local_addr;
    if (packet_type != HCI_EVENT_PACKET) return;

    uint8_t event_type = hci_event_packet_get_type(packet);
    switch(event_type){
        case BTSTACK_EVENT_STATE:
            if (btstack_event_state_get_state(packet) != HCI_STATE_WORKING) return;
            gap_local_bd_addr(local_addr);
            printf("BTstack up and running on %s.\n", bd_addr_to_str(local_addr));
             break;
         case HCI_EVENT_DISCONNECTION_COMPLETE:
            le_notification_enabled = 0;
            printf("Disconnected\n");
            connect_flag = 0;
            break;
        default:
            break;
    }
}

static void nordic_spp_packet_handler(uint8_t packet_type, uint16_t channel, uint8_t *packet, uint16_t size){
    UNUSED(channel);
    switch (packet_type){
        case HCI_EVENT_PACKET:
            if (hci_event_packet_get_type(packet) != HCI_EVENT_GATTSERVICE_META) break;

            switch (hci_event_gattservice_meta_get_subevent_code(packet)){
                case GATTSERVICE_SUBEVENT_SPP_SERVICE_CONNECTED:
                    con_handle = gattservice_subevent_spp_service_connected_get_con_handle(packet);
                    printf("Connected with handle 0x%04x\n", con_handle);
                    connect_flag = 1;
                    break;
                case GATTSERVICE_SUBEVENT_SPP_SERVICE_DISCONNECTED:
                    con_handle = HCI_CON_HANDLE_INVALID;
                    connect_flag = 0;
                     break;
                default:
                    break;
            }
            break;
        case RFCOMM_DATA_PACKET:
            printf("RECV: ");
            printf_hexdump(packet, size);
            break;
        default:
            break;
    }
}

int main() 
{
    // enable RP2040 UART for Serial Monitor
    stdio_init_all();

    // initialize CYW43 driver architecture (will enable BT if/because CYW43_ENABLE_BLUETOOTH == 1)
    if (cyw43_arch_init()) 
    {
        printf("failed to initialise cyw43_arch\n");
        return -1;
    }
    // initialize Logical Link Control and Adaptation Protocol layer
    l2cap_init();
    // initialize Security Manager Protocol layer
    sm_init();
   
    // setup packet handler to inform about BTstack state
    hci_event_callback_registration.callback = &stack_event_packet_handler;
    hci_add_event_handler(&hci_event_callback_registration);
 
    // setup ATT server db attribute database created by compile-gatt.py
    att_server_init(profile_data, NULL,NULL);    
    // register packet handler for ATT server events
    att_server_register_packet_handler(stack_event_packet_handler);

    // setup Nordic SPP service
    nordic_spp_service_server_init(&nordic_spp_packet_handler);

    // setup advertisements
    uint16_t adv_int_min = 800;
    uint16_t adv_int_max = 800;
    uint8_t adv_type = 0;
    bd_addr_t null_addr;
    memset(null_addr, 0, 6);
    gap_advertisements_set_params(adv_int_min, adv_int_max, adv_type, 0, null_addr, 0x07, 0x00);
    assert(adv_data_len <= 31); // ble limitation
    gap_advertisements_set_data(adv_data_len, (uint8_t*) adv_data);
    gap_advertisements_enable(1);
 
    // set one-shot btstack timer
    heartbeat.process = &heartbeat_handler;
    connect_flag = 0;
    btstack_run_loop_set_timer(&heartbeat, HEARTBEAT_PERIOD_MS);
    btstack_run_loop_add_timer(&heartbeat);

    // turn on bluetooth!
    hci_power_control(HCI_POWER_ON);
    
    // btstack_run_loop_execute is only required when using the 'polling' method (e.g. using pico_cyw43_arch_poll library).
    // This example uses the 'threadsafe background` method, where BT work is handled in a low priority IRQ, so it
    // is fine to call bt_stack_run_loop_execute() but equally you can continue executing user code.

#if 0 // btstack_run_loop_execute() is not required, so lets not use it
    btstack_run_loop_execute();
#else
    // this core is free to do it's own stuff except when using 'polling' method (in which case you should use 
    // btstacK_run_loop_ methods to add work to the run loop.
    
    // this is a forever loop in place of where user code would go.
    while(true) {      
        sleep_ms(1000);
    }
#endif
    return 0;
}
